% function [geodesics, lengths, Logs, fails, solutions] = compute_all_geodesics (coeffs, mu, manifold)
%   Compute all geodesics between points in the data set and a specific point 'mu'.
%   This function uses 'parfor' so it can be parallized by starting multiple
%   workers.
%
%   Input:
%     coeffs:   a NxD matrix of D-dimensional data.
%     mu:       a D-vector.
%     manifold: a structure describing how the metric tensor changes. This is
%               created with the 'create_manifold' function.
%
%   Output:
%     geodesics: a N-element cell array containing the geodesics.
%     lengths:   a N-vector containing the lengths of the geodesics.
%     Logs:      a NxD matrix containing the log maps of the data wrt. 'mu'.
%     fails:     a N-element boolean vector used to indicate if the numerical
%                computation converged or not.
%     solutions: a N-element cell array containing the output structures from
%                the ODE solver. This is mostly used for debugging.
%
%   See also:
%     compute_geodesic, create_manifold

%% Copyright 2012 Soren Hauberg
%%
%% This program is free software: you can redistribute it and/or modify
%% it under the terms of the GNU General Public License as published by
%% the Free Software Foundation, either version 3 of the License, or
%% (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program.  If not, see <http://www.gnu.org/licenses/>.

function [geodesics, lengths, Logs, fails, solutions, pga_time] = compute_all_geodesics (coeffs, mu, manifold)
t = tic ();
[N, D] = size (coeffs);
geodesics = cell (N, 1);
lengths = NaN (N, 1);
Logs = NaN (N, D);
fails = false (N, 1);
solutions = cell (N, 1);
parfor n = 1:N
  fprintf ('Iteration %d / %d\n', n, N);
  [curve, len, logarithm, failed, solution] = compute_geodesic (mu, coeffs (n, :), manifold);
  geodesics {n} = curve;
  lengths (n) = len;
  Logs (n, :) = logarithm;
  fails (n) = failed;
  solutions {n} = solution;
end % parfor
pga_time = toc (t);
