%% Copyright (C) 2006-2012, Thomas Treichl <treichl@users.sourceforge.net>
%% Copyright (C) S?ren Hauberg <hauberg@gmail.com>
%%
%% This program is free software; you can redistribute it and/or modify
%% it under the terms of the GNU General Public License as published by
%% the Free Software Foundation; either version 3 of the License, or
%% (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; If not, see <http://www.gnu.org/licenses/>.

%% History:
%%   This function is a modification of 'ode45' from 'odepkg' from the
%%   Octave-Forge project. The original implementation was done by Marc Compere
%%   and released under the GPL. Thomas Treichl then adapted this to GNU Octave
%%   and finally S?ren Hauberg adapted it to the Exponential map use-case.

function retval = myode45 (odefun, ~, v_init, manifold)
  %% XXX: Can we pre-allocate result arrays? 
  %% XXX: Transpose results

  v_init = v_init (:).'; % ensure 'v_init' is a row vector
  dim = numel (v_init) / 2;
  dest_len = norm (v_init (dim+1:end));
  
  %% Options
  trange = [0, 1];
  rel_tol = 1e-6;
  abs_tol = 1e-6;
  stepsize = (trange (2) - trange (1)) / 10;
  max_step = abs (trange (2) - trange (1)) / 10;
  min_step = eps;
  max_iter = 1000;
  
  %% Starting the initialisation of the core solver ode45 
  timestart = trange (1);
  timestop = trange (2);
  
  retval_time = timestart; % first timestamp output
  retval_result = v_init;  % first solution output 
  retval_lengths = 0;
  
  %% Runge-Kutta coefficients (see p.91 in Ascher & Petzold)
  vpow = 1/5;
  va = [0, 0, 0, 0, 0;
        1/4, 0, 0, 0, 0;
        3/32, 9/32, 0, 0, 0;
        1932/2197, -7200/2197, 7296/2197, 0, 0;
        439/216, -8, 3680/513, -845/4104, 0;
        -8/27, 2, -3544/2565, 1859/4104, -11/40];
  
  %% 4th and 5th order b-coefficients
  vb4 = [25/216; 0; 1408/2565; 2197/4104; -1/5; 0];
  vb5 = [16/135; 0; 6656/12825; 28561/56430; -9/50; 2/55];
  vc = sum (va, 2);

  %% The solver main loop - stop if the endpoint has been reached
  cntloop = 2;
  vu = v_init;
  vk = vu.' * zeros (1, 6);
  cntiter = 0;
  cntsave = 2;
  cur_len = 0;
  while ((cur_len < dest_len) && (stepsize >= min_step) && (cntiter <= max_iter))
    %% Estimate the six results when using this solver (XXX: Can we vectorize this?)
    for k = 1:6
      thetime  = timestart + vc (k,1) * stepsize;
      theinput = vu.' + stepsize * vk (:, 1:k-1) * va (k, 1:k-1).';
      vk (:, k) = odefun (thetime, theinput);
    end % for

    %% Compute the 4th and the 5th order estimation
    y4 = vu.' + stepsize * (vk * vb4);
    y5 = vu.' + stepsize * (vk * vb5);
    
    %% Calculate the absolute local truncation error and the acceptable error
    delta = abs (y5 - y4);
    tau = max (rel_tol * abs (vu.'), abs_tol);
    
    %% If the error is acceptable then update the vretval variables
    if (all (delta <= tau))
      timestart = timestart + stepsize;
      vu = y5.';
      
      %% Save the solution            
      retval_time (cntsave, :) = timestart;
      retval_result (cntsave, :) = vu;
      
      %% Update the curve length
      cur_len = curve_length (retval_result (:, 1:dim).', manifold); %% XXX: Do this in a smarter (recursive?) way!
      retval_lengths (cntsave) = cur_len;
      
      %% Update counters
      cntsave = cntsave + 1;    
      cntloop = cntloop + 1;
      cntiter = 0;
    end % if

    %% Update the step size for the next integration step
    delta (delta == 0) = max (delta);
    delta (delta == 0) = max (tau) .* (0.4 ^ (1 / vpow));
    stepsize = min (max_step, min (0.8 * stepsize * (tau ./ delta).^vpow));
      
    %% Update counters that count the number of iteration cycles
    cntiter = cntiter + 1;
  end % while (the main loop)

  %% Check if the curve is too short or too long
  if (cur_len > dest_len)
    retval_result = shorten_curve (retval_lengths, retval_result, dest_len);
  elseif (cur_len < dest_len)
    warning ('could not make geodesic long enough in the given number of iterations');
  end % if
  
  retval.x = retval_time;   % Time stamps are saved in field x
  retval.y = retval_result.'; % Results are saved in field y
end % function

